/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <wba/wba.h>
#include <bld/environ.h>
#include <bld/context.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_Context
***********************************************************************/
const GPL_String BLD_Context::rootScopeName("ROOT");
const GPL_String BLD_Context::scopeDelimiter(".");
GPL_RegularExpression BLD_Context::internalTargetPattern("^__");

BLD_Context::BLD_Context(void)
{
	BLD_Scope rootScope(rootScopeName);
	scopes.Push(rootScope);
}

BLD_Context::~BLD_Context(void)
{
}

GPL_Boolean BLD_Context::AddRule(const GPL_String & string,
									const BLD_RuleType type,
									const OpArray & ops)
{
	// The constructor guarantees at least one scope.
	return scopes.Top().AddRule(string, type, ops);
}

BLD_BuildStatus BLD_Context::Build(const GPL_String & name)
{
	// If one or more rules exist for this target...
	register BLD_BuildStatus status = BLD_BUILD_UNNECESSARY;
	BLD_Target target;
	if (CreateTarget(name, target))
	{
		// Return the value determined by the build.
		status = target.Build();
//WBA_APP->Printf("%s => %d", name.GetData(), status);
	}
	else if (BLD_ARG_ON("BLD_DEBUG"))
	{
		WBA_APP->Printf("Target %s does not exist", name.GetData());
	}
	return status;
}

BLD_BuildStatus BLD_Context::Check(const GPL_String & name,
									const GPL_Boolean descend)
{
//WBA_APP->Printf("%s%s", name.GetData(), BLD_STATE(name).ToString().GetData());
//WBA_APP->Printf("%s%s", reference.GetData(), BLD_STATE(reference).ToString().GetData());

	// If this is a nameless target...
	if (0 == name.GetLength())
	{
		return BLD_BUILD_UNNECESSARY;
	}

	// If this target has previously been checked...
	if (BLD_STATE(name).GetChecked())
	{
		// If this target failed...
		if (BLD_STATE(name).GetFailed())
		{
			// It still fails.
//WBA_APP->Printf("%s%s BLD_BUILD_FAILED", name.GetData(), BLD_STATE(name).ToString().GetData());
			return BLD_BUILD_FAILED;
		}

		// If this target has a time stamp...
		else if (BLD_STATE(name).HasTimeStamp())
		{
			// No further work is necessary.
			return BLD_BUILD_UNNECESSARY;
		}
	}

	// Mark this target as having been checked.
	BLD_STATE(name).SetChecked();

	// If descent is indicated...
	BLD_BuildStatus status = BLD_BUILD_UNNECESSARY;
	if (descend)
	{
		// Return the value determined by the build.
		status = Build(name);
	}

	return status;
}

GPL_Boolean BLD_Context::Pop(BLD_Scope & scope)
{
	return scopes.Pop(scope);
}

GPL_Boolean BLD_Context::Push(const BLD_Scope & scope)
{
	return scopes.Push(scope);
}

void BLD_Context::SetValue(const GPL_String & name, const BLD_Value & value)
{
	// Determine the scope and variable names.
	GPL_String scope;
	GPL_String variable;
	name.Split(scopeDelimiter, scope, variable);
	if (name == scope)
	{
		variable = name;
		scope.Clear();
	}
	SetValue(scope, variable, value);
}

void BLD_Context::SetValue(const GPL_String & scope,
							const GPL_String & variable,
							const BLD_Value & value)
{
	GPL_StackIterator<BLD_Scope> iterator;
	iterator.Reset(&scopes);
	while (iterator.Next())
	{
		if (0 == scope.GetLength())
		{
			iterator.Value().Insert(variable, value);
			break;
		}
		else if (scope == iterator.Value().GetName())
		{
			iterator.Value().Insert(variable, value);
			break;
		}
	}
}

BLD_Value BLD_Context::GetValue(const GPL_String & name) const
{
	// Determine the scope and variable names.
	GPL_String scope;
	GPL_String variable;
	name.Split(scopeDelimiter, scope, variable);
	if (name == scope)
	{
		variable = name;
		scope.Clear();
	}
	return GetValue(scope, variable);
}


BLD_Value BLD_Context::GetValue(const GPL_String & scope,
								const GPL_String & variable) const
{
	BLD_Value value;
	GPL_StackIterator<BLD_Scope> iterator;
	iterator.Reset((GPL_Stack<BLD_Scope> *) &scopes);
	while (iterator.Next())
	{
		if (0 == scope.GetLength())
		{
			if (iterator.Value().LookUp(variable, value))
			{
				return value;
			}
		}
		else if (scope == iterator.Value().GetName())
		{
			iterator.Value().LookUp(variable, value);
			return value;
		}
	}

	// If all else fails, try the environment.
	GPL_String s;
	BLD_Environment::GetEnvironment()->GetValue(variable, s);
	return s;
}

void BLD_Context::GetValues(GPL_Map<GPL_String, BLD_Value> & values) const
{
	for (unsigned long i = 0; scopes.GetCount() > i; ++i)
	{
		scopes[i].GetValues(values);
	}
}

// TODO: Add variable expansion inside back-ticks and support
// for single-tick protection.
GPL_String BLD_Context::Evaluate(const GPL_String & text) const
{
	GPL_String s(text);
	GPL_String v;
	register unsigned long n = s.GetLength();
	register unsigned long i = n;
	register unsigned long j = 0;
	while (n > j)
	{
		switch(s[j])
		{
			case '\\':
				s.Remove(j, 1);
				break;
			case '`':
				// If this is the second of the pair...
				if (i < j)
				{
					const GPL_String command = s.GetSubString(i + 1, j - 1);
					BLD_Environment::GetEnvironment()->Execute(command, v, FALSE);
					s.Remove(i, j - i + 1);
					s.Insert(i, v);
					n = s.GetLength();
					j = i + v.GetLength() - 1;
					i = n;
				}
				else
				{
					i = j;
				}
				break;
			case '(':
				// If this is preceded by a dollar sign...
				if (0 < j && '$' == s[j - 1])
				{
					i = j - 1;
				}
				break;
			case ')':
				// If this is a variable...
				if (i + 1 < j && '(' == s[i + 1])
				{
					v = GetValue(s.GetSubString(i + 2, j - 1)).GetString();
					s.Remove(i, j - i + 1);
					s.Insert(i, v);
					n = s.GetLength();
					j = i + v.GetLength() - 1;
					i = n;
				}
				break;
			default:
				// NOP
				break;
		}
		++j;
	}

	return s;
}

GPL_Boolean BLD_Context::CreateTarget(const GPL_String & name,
										BLD_Target & target)
{
	// Set the name.
	target.SetName(name);

	// Check the rules.
	GPL_Boolean hasBeenFound = FALSE;
	for (unsigned long i = 0; scopes.GetCount() > i; ++i)
	{
		if (scopes[i].CreateTarget(target))
		{
			hasBeenFound = TRUE;
		}
	}
	return hasBeenFound;
}
