/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/environ.h>
#include <bld/fnop.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_EchoOperation
***********************************************************************/
BLD_Value BLD_EchoOperation::Perform(void) const
{
	// If there are any arguments.
	GPL_String output;
	const unsigned long n = ops.GetCount();
	if (0 < n)
	{
		// Build the string.
		GPL_String string("");
		for (unsigned long i = 0; n > i; ++i)
		{
			string.Join(" ", ops[i]->Perform().GetString ());
		}

		// Add a trailing new-line.
		string.Append('\n');

		// Print the string.
		if (!Write(string))
		{
			return BLD_Value::falsehood;
		}
	}

	return BLD_Value::truth;
}

GPL_String BLD_EchoOperation::ToString(void) const
{
	GPL_String image = "echo";
	const unsigned long n = ops.GetCount();
	for (unsigned long i = 0; n > i; ++i)
	{
		image.Join(" ", ops[i]->ToString());
	}
	return image;
}

GPL_Boolean BLD_EchoOperation::Write(const GPL_String & string) const
{
	osdOutputStr(FALSE, string.GetData());

	// osdOutputStr() provides no useful return value.
	return TRUE;
}

/***********************************************************************
class BLD_FileEchoOperation
***********************************************************************/
GPL_String BLD_FileEchoOperation::ToString(void) const
{
	GPL_String image = BLD_EchoOperation::ToString();
	if (append)
	{
		image.Append(" > ");
	}
	else
	{
		image.Append(" >> ");
	}
	image.Append(file->ToString());

	return image;
}

GPL_Boolean BLD_FileEchoOperation::Write(const GPL_String & string) const
{
	BLD_Value value = file->Perform();
	if (0 == value.GetCount())
	{
		return FALSE;
	}

	unsigned long flags = OSD_FILE_WRITE;
	if (!append)
	{
		flags = flags | OSD_FILE_ERASE;
	}

	GPL_File f;
	if (!f.Open(value.GetElement(0), flags))
	{
		return FALSE;
	}

	if (string.GetLength() != f.Write(string.GetData(), string.GetLength()))
	{
		return FALSE;
	}

	f.Close();
	return TRUE;
}

/***********************************************************************
class BLD_SystemOperation
***********************************************************************/
BLD_Value BLD_SystemOperation::Perform(void) const
{
	BLD_Target::SetExecuted();

	// If there are any arguments.
	BLD_Value value;
	if (0 < ops.GetCount())
	{
		// Build the command.
		// NOTE: No shell-protection is performed.  Some Unix arguments
		// may be interpreted differently than expected.
		const unsigned long n = ops.GetCount();
		GPL_String command("");
		for (unsigned long i = 0; n > i; ++i)
		{
			command.Join(" ", ops[i]->Perform().GetString());
		}
		if (BLD_ARG_ON("BLD_VERBATIM"))
		{
			GPL_String string("");
			for (unsigned long i = 0; n > i; ++i)
			{
				string.Join(" ", ops[i]->ToString());
			}
			WBA_APP->Printf("%s", string.GetData());
		}
		if (!quiet && BLD_ARG_OFF("BLD_SILENT") &&
			BLD_CONTEXT.GetValue("SILENT").GetString() != "1")
		{
			WBA_APP->Printf("%s", command.GetData());
		}

		// If execution is *not* disabled...
		if (BLD_ARG_OFF("BLD_NO_EXECUTION"))
		{
			// Execute the command.
			GPL_String output;
			GPL_Boolean inline_print;
			if (!quiet && BLD_ARG_OFF("BLD_QUIET"))
			{
				inline_print = TRUE;
			}
			else
			{
				inline_print = FALSE;
			}
			if (BLD_Environment::GetEnvironment()->Execute
				(command, output, inline_print))
			{
				// Printing output WAS here, but failure needs to print, too.
			}
			else if (BLD_ARG_OFF("BLD_PERSIST") &&
						BLD_ARG_OFF("BLD_IGNORE_SYSTEM_FAILURES") &&
						BLD_CONTEXT.GetValue("IGNORE").GetString() != "1")
			{
				BLD_Target::SetFailed();
			}

			// Parse the output into individual lines.
			GPL_StringTokenizer tokens;
			tokens.Reset (output, "\r\n");
			GPL_String token;
			while ((char *) GPL_BAD_ADDRESS != (token = tokens.Next()).GetData())
			{
				value.AppendElement(token);
			}
		}
	}
	return value;
}

GPL_String BLD_SystemOperation::ToString(void) const
{
	GPL_String image;
	const unsigned long n = ops.GetCount();
	for (unsigned long i = 0; n > i; ++i)
	{
		image.Join(" ", ops[i]->ToString());
	}
	image.Prepend("system '");
	image.Append("'");
	return image;
}

/***********************************************************************
class BLD_ChangeDirectoryOperation
***********************************************************************/
BLD_Value BLD_ChangeDirectoryOperation::Perform(void) const
{
	BLD_Target::SetExecuted();

	const GPL_String d = directory->Perform().GetString();
	if (osdCWDSet(d.GetData()))
	{
		return BLD_Value::truth;
	}
	else
	{
		WBA_APP->Printf("Unable to change to directory %s", d.GetData());
	}

	if (BLD_ARG_OFF("BLD_PERSIST") &&
				BLD_ARG_OFF("BLD_IGNORE_SYSTEM_FAILURES") &&
				BLD_CONTEXT.GetValue("IGNORE").GetString() != "1")
	{
		BLD_Target::SetFailed();
		return BLD_Value::falsehood;
	}

	return BLD_Value::truth;
}

GPL_String BLD_ChangeDirectoryOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("cd %s", directory->ToString().GetData());
	return image;
}

/***********************************************************************
class BLD_IncludeOperation
***********************************************************************/
BLD_Value BLD_IncludeOperation::Perform(void) const
{
	BLD_Target::SetExecuted();

	const GPL_String f = file->Perform().GetString();
	OpArray ops;
	if (BLD_APP->Parse(f, &ops))
	{
		BLD_Target::targets.Top()->IncludeOperations(ops);

		return BLD_Value::truth;
	}
	else
	{
		WBA_APP->Printf("Unable to include %s", f.GetData());
	}

	if (BLD_ARG_OFF("BLD_PERSIST") &&
				BLD_ARG_OFF("BLD_IGNORE_SYSTEM_FAILURES") &&
				BLD_CONTEXT.GetValue("IGNORE").GetString() != "1")
	{
		BLD_Target::SetFailed();
		return BLD_Value::falsehood;
	}

	return BLD_Value::truth;
}

GPL_String BLD_IncludeOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("include %s", file->ToString().GetData());
	return image;
}

/***********************************************************************
class BLD_ExistOperation
***********************************************************************/
BLD_Value BLD_ExistOperation::Perform(void) const
{
	const GPL_String fullPathName = file->Perform().GetString().GetData();
	if (GPL_File::Exists(fullPathName))
	{
		OSD_Time t;
		if (GPL_File::GetModificationTime(fullPathName, t))
		{
			BLD_STATE(fullPathName).SetLastBuilt(t);
		}
		return BLD_Value::truth;
	}
	return BLD_Value::falsehood;
}

GPL_String BLD_ExistOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("exists(%s)", file->ToString().GetData());;
	return image;
}

/***********************************************************************
class BLD_DateOperation
***********************************************************************/
BLD_Value BLD_DateOperation::Perform(void) const
{
	OSD_Time t;
	GPL_String image;
	if (GPL_File::GetModificationTime(file->Perform().GetString().GetData(), t))
	{
		const GPL_Time time = t;
		image = time.ToString();
	}
	return image;
}

GPL_String BLD_DateOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("date(%s)", file->ToString().GetData());;
	return image;
}

/***********************************************************************
class BLD_DateSortOperation
***********************************************************************/
BLD_Value BLD_DateSortOperation::Perform(void) const
{
	// For each constituent...
	OSD_Time t;
	GPL_Time time;
	GPL_Time lastWrite;
	GPL_Boolean first = TRUE;
	const unsigned long n = ops.GetCount();
	for (unsigned long i = 0; n > i; ++i)
	{
		// Get its modification time.
		if (GPL_File::GetModificationTime(ops[i]->Perform().GetString().GetData(), t))
		{
			time = t;
			// If this is the first time through...
			if (first)
			{
				lastWrite = time;
				first = FALSE;
			}

			// If this is not the first time through...
			else if ((!ascending && time > lastWrite) &&
					(ascending && time < lastWrite))
			{
				lastWrite = time;
			}
		}
	}
	return lastWrite.ToString();
}

GPL_String BLD_DateSortOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("date(%s)", (0 == ops.GetCount() ? "" : ops.operator[](0)->ToString().GetData()));
	return image;
}


/***********************************************************************
class BLD_SpecificationOperation
***********************************************************************/
BLD_Value BLD_SpecificationOperation::Perform(void) const
{
	BLD_CONTEXT.AddRule(target, type, ops);
	return BLD_Value::truth;
}

GPL_String BLD_SpecificationOperation::ToString(void) const
{
	GPL_String image;
	image.Append(target);
	switch(type)
	{
	case BLD_RULE_FAIL_OPERATIONS:
		image.Append(":x");
		break;

	case BLD_RULE_PREPEND_OPERATIONS:
		image.Append(":-");
		break;

	case BLD_RULE_APPEND_OPERATIONS:
		image.Append(":+");
		break;

	default:
		image.Append(":");
		break;
	}
	image.Join("\n", "{");
	for (unsigned long i = 0; ops.GetCount() > i; ++i)
	{
		image.Join("\n", ops[i]->ToString());
	}
	image.Join("\n", "}");
	return image;
}

